;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "org-project-capture" "20260313.1738"
  "Repository todo capture and management for org-mode."
  '((dash                 "2.10.0")
    (emacs                "28")
    (s                    "1.9.0")
    (org-category-capture "1.0.0"))
  :url "https://github.com/colonelpanic8/org-project-capture"
  :commit "6a95fb90bcdb7fcdeba9d9421d7c511cea95ef70"
  :revdesc "6a95fb90bcdb"
  :keywords '("org-mode" "todo" "tools" "outlines" "project" "capture")
  :authors '(("Ivan Malison" . "IvanMalison@gmail.com"))
  :maintainers '(("Ivan Malison" . "IvanMalison@gmail.com")))
